open import Level using (Level)
open import Function using (_∘_; _$_; flip) 
open import Relation.Nullary.Core using (¬_; yes; no)
open import Relation.Binary using 
                     (Rel; Setoid; module Setoid; 
                      Decidable; DecSetoid; module DecSetoid;
                      IsDecEquivalence; module IsDecEquivalence)
open import Relation.Binary.Core using 
                                (_≡_; Reflexive; Symmetric; Transitive; 
                                    IsEquivalence; module IsEquivalence)
open import Data.Product using (_×_; _,_; proj₁; proj₂)


productSetoid : {c l c' l' : Level} -> (A : Setoid c  l ) -> 
                                       (B : Setoid c' l') -> Setoid _ _
-- Setoid  for  Carrier A × Carrier B.

productSetoid A B =  record{ Carrier        = Carr-A × Carr-B
                            ; _≈_           = _=p_
                            ; isEquivalence = isEquiv-Product }
  where
  open Setoid A renaming (Carrier to Carr-A; _≈_ to _=a_; 
                                               isEquivalence to isEq-A)
  open Setoid B renaming (Carrier to Carr-B; _≈_ to _=b_; 
                                               isEquivalence to isEq-B)
  open IsEquivalence isEq-A renaming 
                        (refl to refl-A; sym to sym-A; trans to trans-A)
  open IsEquivalence isEq-B renaming 
                        (refl to refl-B; sym to sym-B; trans to trans-B)

  _=p_ : Rel (Carr-A × Carr-B) _
  pr =p pr' =  (proj₁ pr =a proj₁ pr') × (proj₂ pr =b proj₂ pr')

  isEquiv-Product : IsEquivalence _=p_     
  isEquiv-Product = record{ refl  = (λ {p}     -> refl-P  {p})  
                          ; sym   = (λ {p q}   -> sym-P   {p} {q})  
                          ; trans = (λ {p q r} -> trans-P {p} {q} {r}) }
    where
    refl-P : Reflexive _=p_
    refl-P = (refl-A , refl-B)

    sym-P : Symmetric _=p_             -- pair =p pair' -> pair' =p pair
    sym-P pr=pr' =  (sym-A a=a' , sym-B b=b') where  a=a' = proj₁ pr=pr'
                                                     b=b' = proj₂ pr=pr'
    trans-P : Transitive _=p_  
    trans-P pr=pr' pr'=pr'' =  
                             (trans-A a=a' a'=a'' , trans-B b=b' b'=b'')
                             where
                             a=a'   = proj₁ pr=pr'
                             a'=a'' = proj₁ pr'=pr''
                             b=b'   = proj₂ pr=pr'
                             b'=b'' = proj₂ pr'=pr''

------------------------------------------------------------------------
setoidByMap : {c c' l' : Level} -> (C : Set c) -> (B : Setoid c' l') ->
                                   (C -> Setoid.Carrier B) -> Setoid _ _
setoidByMap C B f = 
          record{ Carrier = C;  _≈_ = _=a_;  isEquivalence = isEquiv-A }
          where
          open Setoid {{...}}
          _=a_ : Rel C _
          a =a a' =  f a ≈ f a' 

          isEquiv-A : IsEquivalence _=a_
          isEquiv-A = record { refl = refl; sym = sym; trans = trans}